import requests
from .HTTPError import HTTPError

class API: 
    def __init__(self, host):
        if host[-1:] != "/":
            host = host + "/"

        self.host = host
        self.headers = {}

    def set_header(self, header, value):
        self.headers[header] = value
        return self

    def set_headers(self, headers):
        self.headers = headers
        return self

    def post(self, endpoint, json = {}):
        r = requests.post(self.host + endpoint, headers=self.headers, json=json) 

        if r.status_code > 299:
            raise HTTPError(r.json())

        if len(r.content) == 0:
            return r
            
        return r.json()

    def get(self, endpoint):
        r = requests.get(self.host + endpoint, headers=self.headers)

        if r.status_code > 299:
            raise HTTPError(r.json())

        if len(r.content) == 0:
            return r
            
        return r.json()

    def delete(self, endpoint):
        r = requests.delete(self.host + endpoint, headers=self.headers)

        if r.status_code > 299:
            raise HTTPError(r.json())

        if len(r.content) == 0:
            return r
            
        return r.json()


import requests
import time
from .API import API
from .services import UserService


class Session:
    def __init__(self, api, api_key, version="1.0"):
        self.__api_key = api_key
        self.validity = None
        self.refreshed = None
        self.__token = None
        self.headers = None
        self.api = api
        self.version = version

        self.getToken()

    def getToken(self):
        if self.validity != None and time.gmtime() - self.refreshed < self.validity:
            return self.__token

        if self.version != "1.0" and self.version != "2.0":
            print('Provided version not available')

        r = self.api.set_header("Authorization", "Basic " + self.__api_key) \
                .set_header("basiq-version", self.version) \
                .post("token", {})

        if "access_token" in r:
            self.refreshed = time.gmtime()
            self.validity = r["expires_in"]
            self.__token = r["access_token"]
            self.api.headers = {
                "Authorization": "Bearer " + r["access_token"]
            }
            return r["access_token"]
        else:
            print("No access token:", r)

    def getInstitutions(self):
        return self.api.get("institutions")

    def getInstitution(self, id):
        return self.api.get("institutions/" + id)

    def getUser(self, id):
        return UserService(self).get(id)

    def forUser(self, id):
        return UserService(self).forUser(id)



from ..utils import FilterBuilder
from .ConnectionService import ConnectionService

class TransactionList:
    def __init__(self, service, data):
        self.data = data
        self.service = service

    def next(self): 
        if "links" in self.data and "next" in self.data["links"]:
            next_link = self.data["links"]["next"]
            print("Next:", next_link)
            next_path = next_link[next_link.rfind(".io/")+4:]
            nt = self.service.session.api.get(next_path)
            self.data = nt
            return self

        return None

class User:
    def __init__(self, service):
        self.service = service

    def forUser(self, id):
        self.id = id
        return self

    def get(self):
        r = self.service.get(self.id)
        self.email = r["email"]
        self.mobile = r["mobile"]
        return self

    def update(self, email = None, mobile = None):
        return self.service.update(self.id, email=email, mobile=mobile)

    def delete(self):
        return self.service.delete(self.id)

    def createConnection(self, data):
        return ConnectionService(self.service.session, self).create(data)

    def refreshAllConnections(self):
        return self.service.refreshAllConnections(self.id)

    def listAllConnections(self, filter=None):
        return self.service.listAllConnections(self.id, filter)

    def getTransaction(self, id):
        return self.service.getTransaction(self.id, id)

    def getTransactions(self, filter = None):
        return self.service.getTransactions(self.id, filter)

    def getAccount(self, id):
        return self.service.getAccount(self.id, id)

    def getAccounts(self, filter = None):
        return self.service.getAccounts(self.id, filter)


class UserService:
    def __init__(self, session):
        self.session = session
    
    def forUser(self, id):
        return User(self).forUser(id)

    def get(self, id):
        r = self.session.api.get("users/" + id)

        u = User(self)
        u.id = r["id"]
        u.email = r["email"]
        u.mobile = r["mobile"]

        return u

    def create(self, email = None, mobile = None):
        json = {}
        if email == None and mobile == None:
            raise Exception("Neither email or mobile were provided")
        if email != None:
            json["email"] = email
        if mobile != None:
            json["mobile"] = mobile
            
        r = self.session.api.post("users/", json=json)

        u = User(self)
        u.id = r["id"]
        u.email = r["email"]
        u.mobile = r["mobile"]

        return u

    def update(self, id, email = None, mobile = None):
        json = {}
        if email == None and mobile == None:
            raise Exception("Neither email or mobile were provided")
        if email != None:
            json["email"] = email
        if mobile != None:
            json["mobile"] = mobile
            
        r = self.session.api.post("users/" + id, json=json)

        u = User(self)
        u.id = r["id"]
        u.email = r["email"]
        u.mobile = r["mobile"]

        return u

    def delete(self, id):            
        self.session.api.delete("users/" + id)

        return None

    def refreshAllConnections(self, id):
        r = self.session.api.post("users/" + id + "/connections/refresh")

        return r

    def listAllConnections(self, id, filter=None):
        url = "users/" + id + "/connections"
        if filter != None:
            if type(filter).__name__ != "FilterBuilder":
                raise Exception("Provided filter must be an instance of FilterBuilder class")
            url = url + "?" + filter.getFilter()
        return self.session.api.get(url)

    def getTransaction(self, user_id, id):
        return self.session.api.get("users/" + user_id + "/transactions/" + id)

    def getTransactions(self, user_id, filter = None):
        url = "users/" + user_id + "/transactions"
        if filter != None:
            if type(filter).__name__ != "FilterBuilder":
                raise Exception("Provided filter must be an instance of FilterBuilder class")
            url = url + "?" + filter.getFilter()

        t = self.session.api.get(url)

        return TransactionList(self, t)
        
    def getAccount(self, user_id, id):
        return self.session.api.get("users/" + user_id + "/accounts/" + id)

    def getAccounts(self, user_id, filter = None):
        url = "users/" + user_id + "/accounts"
        if filter != None:
            if type(filter).__name__ != "FilterBuilder":
                raise Exception("Provided filter must be an instance of FilterBuilder class")
            url = url + "?" + filter.getFilter()

        return self.session.api.get(url)



import time

class Connection:
    def __init__(self, service):
        self.service = service

    def forConnection(self, id):
        self.id = id
        return self

    def refresh(self):
        return self.service.refresh(self.id)

    def update(self, password):
        return self.service.update(self.id, password)

    def delete(self):
        return self.service.delete(self.id)

class Job:
    def __init__(self, service):
        self.service = service

    def forJob(self, id):
        self.id = id
        return self

    def getConnectionId(self):
        if hasattr(self, "links") and "source" not in self.links:
            return ""
        
        if hasattr(self, "links") == False:
            return ""

        return self.links["source"][self.links["source"].rfind("/")+1:]

    def getConnection(self):
        if self.getConnectionId() == "":
            j = self.service.getJob(self.id)
        else:
            j = self

        return self.service.get(j.getConnectionId())

    def waitForCredentials(self, interval, timeout, i = 0):
        j = self.service.getJob(self.id)

        time.sleep(interval / 1000)

        if i * (interval / 1000) > timeout:
            return False

        i += 1

        step = j.steps[0]
        if step["status"] == "success":
            return self.service.get(j.getConnectionId())
        if step["status"] == "failed":
            return False
            
        return self.waitForCredentials(interval, timeout, i)

    def waitForAccounts(self, interval, timeout, i = 0):
        j = self.service.getJob(self.id)

        time.sleep(interval / 1000)

        if i * (interval / 1000) > timeout:
            return False

        i += 1

        step = j.steps[1]
        if step["status"] == "success":
            return self.service.get(j.getConnectionId())
        if step["status"] == "failed":
            return False

        return self.waitForAccounts(interval, timeout, i)

    def waitForTransactions(self, interval, timeout, i = 0):
        j = self.service.getJob(self.id)

        time.sleep(interval / 1000)

        if i * (interval / 1000) > timeout:
            return False

        i += 1

        step = j.steps[2]
        if step["status"] == "success":
            return self.service.get(j.getConnectionId())
        if step["status"] == "failed":
            return False
            
        return self.waitForTransactions(interval, timeout, i)

class ConnectionService:
    def __init__(self, session, user):
        self.session = session
        self.user = user

    def forConnection(self, id):
        return Connection(self).forConnection(id)

    def get(self, id):
        r = self.session.api.get("users/" + self.user.id + "/connections/" + id)

        c = Connection(self)
        c.id = r["id"]
        c.status = r["status"]
        c.last_used = r["lastUsed"]
        c.institution = r["institution"]
        c.accounts = r["accounts"]
        c.links = r["links"]

        return c

    def getJob(self, id):
        r = self.session.api.get("jobs/" + id)

        j = Job(self)
        j.id = r["id"]
        j.created = r["created"]
        j.updated = r["updated"]
        j.links = r["links"]
        if "steps" in r:
            j.steps = r["steps"]

        return j

    def create(self, connection_data):
        if "loginId" not in connection_data:
            raise Exception("Login id needs to be suplied")
        if "password" not in connection_data:
            raise Exception("Password id needs to be suplied")
        if "institution" not in connection_data:
            raise Exception("Institution data needs to be suplied")
        if "id" not in connection_data["institution"]:
            raise Exception("Institution id needs to be suplied")

        r = self.session.api.post("users/" + self.user.id + "/connections", json=connection_data)

        j = Job(self)
        j.id = r["id"]

        return j

    def update(self, id, password, security_code=None, secondary_login_id=None):
        data = {'password': password}
        if security_code is not None:
            data['securityCode'] = security_code
        if secondary_login_id is not None:
            data['secondaryLoginId'] = secondary_login_id
        r = self.session.api.post("users/" + self.user.id + "/connections/" + id, json=data)

        j = Job(self)
        j.id = r["id"]

        return j

    def delete(self, id):
        self.session.api.delete("users/" + self.user.id + "/connections/" + id)

        return None

    def refresh(self, id):
        r = self.session.api.post("users/" + self.user.id + "/connections/" + id + "/refresh")

        j = Job(self)
        j.id = r["id"]

        return j



class HTTPError(Exception):
    def __init__(self, response):
        self.correlationId = response["correlationId"]
        self.data = response["data"]
        messages = []
        for message in self.data:
            messages.append(message["detail"])
        
        self.msg = ", ".join(messages)

    def get_message(self):
        return self.msg

    def __str__(self):
        return "HTTPError: %s" % self.msg


class FilterBuilder:
    def __init__(self, filters = []):
        self.filters = filters
    
    def eq(self, field, value):
        self.filters.append(field + ".eq('" + value + "')")
        return self
    
    def gt(self, field, value):
        self.filters.append(field + ".gt('" + value + "')")
        return self
    
    def gteq(self, field, value):
        self.filters.append(field + ".gteq('" + value + "')")
        return self
    
    def lt(self, field, value):
        self.filters.append(field + ".lt('" + value + "')")
        return self
    
    def lteq(self, field, value):
        self.filters.append(field + ".lteq('" + value + "')")
        return self
    
    def bt(self, field, valueOne, valueTwo):
        self.filters.append(field + ".bt('" + valueOne + "','" + valueTwo + "')")
        return self
    
    def toString(self):
        return ",".join(self.filters)
    
    def getFilter(self):
        return "filter=" + ",".join(self.filters)
    
    def setFilter(self, filters):
        self.filters = filters
        return self


from setuptools import setup
setup(
    name="basiq",
    packages=["basiq/services", "basiq/utils", "basiq"],
    version="1.0.1",
    description="SDK Package for Basiq's HTTP API",
    author="Nenad Lukic",
    python_requires=">=3",
    install_requires=["requests"],
    author_email="nenad@basiq.io",
    url="https://github.com/basiqio/basiq-sdk-python",
    download_url="https://github.com/basiqio/basiq-sdk-python/archive/1.0.1.tar.gz",
    keywords=["basiq", "finance", "sdk", "api"],
    classifiers=[]
)
